@@ -132,10 +132,12 @@
     Updated by      Richard Gooch   12-APR-2000: More efficient discarding of
   lookups on "/dev/log" and "/dev/initctl".
 
-    Last updated by Richard Gooch   16-APR-2000: Added "MFUNCTION" and
+    Updated by      Richard Gooch   16-APR-2000: Added "MFUNCTION" and
   "CFUNCTION" actions. Added compatibility entries for Computone Multiport
   serial devices.
 
+    Last updated by Richard Gooch   17-APR-2000: Added "COPY" action.
+
 
 */
 #include <unistd.h>
@@ -150,6 +152,8 @@
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <sys/ioctl.h>
+#include <sys/socket.h>
+#include <sys/un.h>
 #include <dirent.h>
 #include <fcntl.h>
 #include <linux/devfs_fs.h>
@@ -170,7 +174,7 @@
 #endif
 
 
-#define VERSION "1.3.3"
+#define VERSION "1.3.4"
 #define CONFIG_FILE "/etc/devfsd.conf"
 #define MAX_ARGS (6 + 1)
 
@@ -188,11 +192,12 @@
 #define AC_EXECUTE                  2
 #define AC_MFUNCTION                3
 #define AC_CFUNCTION                4
-#define AC_IGNORE                   5
-#define AC_MKOLDCOMPAT              6
-#define AC_MKNEWCOMPAT              7
-#define AC_RMOLDCOMPAT              8
-#define AC_RMNEWCOMPAT              9
+#define AC_COPY                     5
+#define AC_IGNORE                   6
+#define AC_MKOLDCOMPAT              7
+#define AC_MKNEWCOMPAT              8
+#define AC_RMOLDCOMPAT              9
+#define AC_RMNEWCOMPAT              10
 
 #define MKACTION(what,when) (struct action_type) {what, when}
 
@@ -225,6 +230,12 @@ struct call_function_type
     char *argv[MAX_ARGS + 1];  /*  argv[0] is the function name          */
 };
 
+struct copy_type
+{
+    const char *source;
+    const char *destination;
+};
+
 struct action_type
 {
     unsigned int what;
@@ -240,6 +251,7 @@ struct config_entry_struct
 	struct permissions_type permissions;
 	struct execute_type execute;
 	struct call_function_type function;
+	struct copy_type copy;
     }
     u;
     struct config_entry_struct *next;
@@ -282,6 +294,8 @@ static void action_execute (const struct devfsd_notify_struct *info,
 			    const struct config_entry_struct *entry);
 static void action_call_function (const struct devfsd_notify_struct *info,
 				  const struct config_entry_struct *entry);
+static void action_copy (const struct devfsd_notify_struct *info,
+			 const struct config_entry_struct *entry);
 static void action_compat (const struct devfsd_notify_struct *info,
 			   unsigned int action);
 static void free_config ();
@@ -294,6 +308,7 @@ static CONST char *get_variable (CONST char *variable, void *info);
 static void do_open_syslog ();
 static void do_scan_and_service (DIR *dp, CONST char *dirname);
 static flag make_symlink (const char *oldpath, const char *newpath);
+static flag make_dir_tree (const char *path);
 static struct shared_object *get_shared_object (CONST char *name);
 
 
@@ -613,14 +628,12 @@ static void read_config (CONST char *filename, flag optional,
 			"PLEASE USE MODLOAD ACTION INSTEAD OF EXECUTE modprobe\n");
 	    for (count = 0; count < num_args; ++count)
 	    {
-		err = strlen (p[count]) + 1;
-		if ( ( new->u.execute.argv[count] = malloc (err) ) == NULL )
+		if ( ( new->u.execute.argv[count] =strdup (p[count]) ) ==NULL )
 		{
 		    SYSLOG (LOG_ERR, "error allocating\n");
 		    SYSLOG (LOG_ERR, "exiting\n");
 		    exit (1);
 		}
-		strcpy (new->u.execute.argv[count], p[count]);
 	    }
 	    new->u.execute.argv[num_args] = NULL;
 	}
@@ -652,17 +665,38 @@ static void read_config (CONST char *filename, flag optional,
 	    --num_args;
 	    for (count = 0; count < num_args; ++count)
 	    {
-		err = strlen (p[count + 1]) + 1;
-		if ( ( new->u.function.argv[count] = malloc (err) ) == NULL )
+		if ( ( new->u.function.argv[count] = strdup (p[count + 1]) )
+		     == NULL )
 		{
 		    SYSLOG (LOG_ERR, "error allocating\n");
 		    SYSLOG (LOG_ERR, "exiting\n");
 		    exit (1);
 		}
-		strcpy (new->u.function.argv[count], p[count + 1]);
 	    }
 	    new->u.function.argv[num_args] = NULL;
 	}
+	else if (strcasecmp (what, "COPY") == 0)
+	{
+	    new->action.what = AC_COPY;
+	    num_args -= 3;
+	    if (num_args != 2)
+	    {
+		SYSLOG (LOG_ERR, "missing path and function in \"%s\"\n",line);
+		exit (1);
+	    }
+	    if ( ( new->u.copy.source = strdup (p[0]) ) == NULL )
+	    {
+		SYSLOG (LOG_ERR, "error allocating source string\n");
+		SYSLOG (LOG_ERR, "exiting\n");
+		exit (1);
+	    }
+	    if ( ( new->u.copy.destination = strdup (p[1]) ) == NULL )
+	    {
+		SYSLOG (LOG_ERR, "error allocating destination string\n");
+		SYSLOG (LOG_ERR, "exiting\n");
+		exit (1);
+	    }
+	}
 	else if (strcasecmp (what, "IGNORE") == 0)
 	    new->action.what = AC_IGNORE;
 	else if (strcasecmp (what, "MKOLDCOMPAT") == 0)
@@ -789,6 +823,9 @@ static void service_name (const struct devfsd_notify_struct *info)
 	  case AC_CFUNCTION:
 	    action_call_function (info, entry);
 	    break;
+	  case AC_COPY:
+	    action_copy (info, entry);
+	    break;
 	  case AC_IGNORE:
 	    return;
 	    /*break;*/
@@ -1017,6 +1054,113 @@ static void action_call_function (const struct devfsd_notify_struct *info,
 	    entry->u.function.argv[0], argv[0]);
 }   /*  End Function action_call_function  */
 
+static void action_copy (const struct devfsd_notify_struct *info,
+			 const struct config_entry_struct *entry)
+/*  [SUMMARY] Copy permissions.
+    <info> The devfs change.
+    <entry> The config file entry.
+    [RETURNS] Nothing.
+*/
+{
+    struct get_variable_info gv_info;
+    struct stat source_stat, dest_stat;
+    char source[STRING_LENGTH], destination[STRING_LENGTH];
+
+    gv_info.info = info;
+    gv_info.devname = info->devname;
+    strcpy (gv_info.devpath, mount_point);
+    strcat (gv_info.devpath, "/");
+    strcat (gv_info.devpath, info->devname);
+    st_expr_expand (source, STRING_LENGTH, entry->u.copy.source, get_variable,
+		    &gv_info, NULL);
+    st_expr_expand (destination, STRING_LENGTH, entry->u.copy.destination,
+		    get_variable, &gv_info, NULL);
+    if (trace_level > 1)
+	fprintf (stderr, "Copying %s to %s for \"%s\" | \"%s\"\n",
+		 source, destination, info->devname, gv_info.devpath);
+    if ( !make_dir_tree (destination) )
+    {
+	SYSLOG (LOG_ERR, "error making tree for: \"%s\"\n", destination);
+	return;
+    }
+    if (lstat (source, &source_stat) != 0) return;
+    if (lstat (destination, &dest_stat) != 0) dest_stat.st_mode = 0;
+    if ( (source_stat.st_mode & S_IFMT) == (dest_stat.st_mode & S_IFMT) )
+    {
+	/*  Same type  */
+	if ( S_ISLNK (source_stat.st_mode) )
+	{
+	    int source_len, dest_len;
+	    char source_link[STRING_LENGTH], dest_link[STRING_LENGTH];
+
+	    source_len = readlink (source, source_link, STRING_LENGTH - 1);
+	    source_link[source_len] = '\0';
+	    dest_len = readlink (destination, dest_link, STRING_LENGTH - 1);
+	    dest_link[dest_len] = '\0';
+	    if ( (source_len != dest_len) ||
+		 (strcmp (source_link, dest_link) != 0) )
+	    {
+		unlink (destination);
+		symlink (source_link, destination);
+	    }
+	    return;
+	}
+	if (source_stat.st_mode != dest_stat.st_mode)
+	    chmod (destination, source_stat.st_mode & ~S_IFMT);
+	if ( (source_stat.st_uid != dest_stat.st_uid) ||
+	     (source_stat.st_gid != dest_stat.st_gid) )
+	    chown (destination, source_stat.st_uid, source_stat.st_gid);
+	return;
+    }
+    /*  Different types: unlink and create  */
+    unlink (destination);
+    switch (source_stat.st_mode & S_IFMT)
+    {
+	int fd, val;
+	struct sockaddr_un un_addr;
+	char symlink_val[STRING_LENGTH];
+
+      case S_IFSOCK:
+	if ( ( fd = socket (AF_UNIX, SOCK_STREAM, 0) ) < 0 ) break;
+	un_addr.sun_family = AF_UNIX;
+	strcpy (un_addr.sun_path, destination);
+	if (bind (fd, (struct sockaddr *) &un_addr, (int) sizeof un_addr) != 0)
+	    break;
+	if (chmod (destination, source_stat.st_mode & ~S_IFMT) != 0) break;
+	if (chown (destination, source_stat.st_uid, source_stat.st_gid) == 0)
+	    return;
+	break;
+      case S_IFLNK:
+	if ( ( val = readlink (source, symlink_val, STRING_LENGTH - 1) ) < 0 )
+	    break;
+	symlink_val[val] = '\0';
+	if (symlink (symlink_val, destination) == 0) return;
+	break;
+      case S_IFREG:
+	if (open (destination, O_RDONLY | O_CREAT,
+		  source_stat.st_mode & ~S_IFMT) < 0) break;
+	if (chmod (destination, source_stat.st_mode & ~S_IFMT) != 0) break;
+	if (chown (destination, source_stat.st_uid, source_stat.st_gid) == 0)
+	    return;
+	break;
+      case S_IFBLK:
+      case S_IFCHR:
+      case S_IFIFO:
+	if (mknod (destination, source_stat.st_mode, source_stat.st_rdev) != 0)
+	    break;
+	if (chown (destination, source_stat.st_uid, source_stat.st_gid) == 0)
+	    return;
+	break;
+      case S_IFDIR:
+	if (mkdir (destination, source_stat.st_mode & ~S_IFMT) != 0) break;
+	if (chown (destination, source_stat.st_uid, source_stat.st_gid) == 0)
+	    return;
+	return;
+	/*break;*/
+    }
+    SYSLOG (LOG_ERR, "error copying: \"%s\" to \"%s\"\n", source, destination);
+}   /*  End Function action_copy  */
+
 static void action_compat (const struct devfsd_notify_struct *info,
 			   unsigned int action)
 /*  [SUMMARY] Process a compatibility request.
@@ -1772,35 +1916,9 @@ static flag make_symlink (const char *oldpath, const char *newpath)
     [RETURNS] TRUE on success, else FALSE.
 */
 {
-    char ch;
-    char *ptr1, *ptr2;
-    char newpath_copy[STRING_LENGTH];
     static char function_name[] = "make_symlink";
 
-    strcpy (newpath_copy, newpath);
-    ptr1 = newpath_copy;
-    while (TRUE)
-    {
-	if ( ( ptr2 = strchr (ptr1 + 1, '/') ) == NULL ) break;
-	ch = *ptr2;
-	*ptr2 = '\0';
-	if (access (newpath_copy, F_OK) == 0)
-	{
-	    *ptr2 = ch;
-	    ptr1 = ptr2;
-	    continue;
-	}
-	if (mkdir (newpath_copy,
-		   S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) != 0)
-	{
-	    SYSLOG (LOG_ERR, "Error making directory \"%s\"\t%s\n",
-		    newpath_copy, ERRSTRING);
-	    *ptr2 = ch;
-	    return (FALSE);
-	}
-	*ptr2 = ch;
-	ptr1 = ptr2;
-    }
+    if ( !make_dir_tree (newpath) ) return (FALSE);
     if (symlink (oldpath, newpath) != 0)
     {
 	if (errno == EEXIST)
@@ -1816,6 +1934,44 @@ static flag make_symlink (const char *oldpath, const char *newpath)
     return (TRUE);
 }   /*  End Function make_symlink  */
 
+static flag make_dir_tree (const char *path)
+/*  [SUMMARY] Creating intervening directories for a path as required.
+    <path> The full pathname (including he leaf node).
+    [RETURNS] TRUE on success, else FALSE.
+*/
+{
+    char ch;
+    char *ptr1, *ptr2;
+    char path_copy[STRING_LENGTH];
+    /*static char function_name[] = "make_dir_tree";*/
+
+    strcpy (path_copy, path);
+    ptr1 = path_copy;
+    while (TRUE)
+    {
+	if ( ( ptr2 = strchr (ptr1 + 1, '/') ) == NULL ) break;
+	ch = *ptr2;
+	*ptr2 = '\0';
+	if (access (path_copy, F_OK) == 0)
+	{
+	    *ptr2 = ch;
+	    ptr1 = ptr2;
+	    continue;
+	}
+	if (mkdir (path_copy,
+		   S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) != 0)
+	{
+	    SYSLOG (LOG_ERR, "Error making directory \"%s\"\t%s\n",
+		    path_copy, ERRSTRING);
+	    *ptr2 = ch;
+	    return (FALSE);
+	}
+	*ptr2 = ch;
+	ptr1 = ptr2;
+    }
+    return (TRUE);
+}   /*  End Function make_dir_tree  */
+
 static struct shared_object *get_shared_object (CONST char *name)
 /*  [SUMMARY] Load a shared object if possible.
     <name> The shared object.
